import { Injectable, ɵɵdefineInjectable, ɵɵinject } from '@angular/core';
import { NavigationEnd, Router } from '@angular/router';
import { identityMapper, deserializeData, serializeData } from '@softeq/data-mappers';
import { Subscription, BehaviorSubject, combineLatest, of, from } from 'rxjs';
import assign from 'lodash/assign';
import isEqual from 'lodash/isEqual';
import isNil from 'lodash/isNil';
import { map, filter, tap, switchMap, skip, startWith } from 'rxjs/operators';
import flatMap from 'lodash/flatMap';
import groupBy from 'lodash/groupBy';
import mapValues from 'lodash/mapValues';
import toPairs from 'lodash/toPairs';
import { FormControl } from '@angular/forms';
import create from 'lodash/create';
import every from 'lodash/every';
import fromPairs from 'lodash/fromPairs';
import isEmpty from 'lodash/isEmpty';
import isString from 'lodash/isString';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const URL_RE = /^([^?#]*)(?:\?([^#]*))?(?:#(.*))?$/;
/**
 * @param {?} url
 * @return {?}
 */
function parseUrl(url) {
    /** @type {?} */
    const match = URL_RE.exec(url);
    if (match) {
        const [_, pathname, query, hash] = match;
        return {
            pathname,
            hash,
            query: mapValues(mapValues(groupBy((query || '')
                .split('&')
                .map((/**
             * @param {?} param
             * @return {?}
             */
            (param) => param.split('=').map(decodeURIComponent))), (/**
             * @param {?} __0
             * @return {?}
             */
            ([key]) => key)), (/**
             * @param {?} pairs
             * @return {?}
             */
            (pairs) => pairs.map((/**
             * @param {?} __0
             * @return {?}
             */
            ([_, value]) => value)))), (/**
             * @param {?} values
             * @return {?}
             */
            (values) => values.length === 1 ? values[0] : values)),
        };
    }
    else {
        return { pathname: '', query: {}, hash: '' };
    }
}
/**
 * @param {?} url
 * @return {?}
 */
function stringifyUrl(url) {
    return [
        url.pathname,
        url.query ? '?' : '',
        flatMap(toPairs(url.query)
            .filter((/**
         * @param {?} __0
         * @return {?}
         */
        ([_, value]) => !isNil(value)))
            .map((/**
         * @param {?} __0
         * @return {?}
         */
        ([key, value]) => [key, Array.isArray(value) ? value : [value]])), (/**
         * @param {?} __0
         * @return {?}
         */
        ([key, values]) => values.map((/**
         * @param {?} value
         * @return {?}
         */
        (value) => `${encodeURIComponent(key)}=${encodeURIComponent((value))}`)))).join('&'),
        url.hash ? '#' : '',
        url.hash,
    ].join('');
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Checks if provided snapshot is in the snapshot tree of the router state
 * @param {?} router
 * @param {?} snapshot
 * @return {?}
 */
function hasRouteInState(router, snapshot) {
    /** @type {?} */
    const snapshotStack = [];
    snapshotStack.push(router.snapshot.root);
    // simple tree traversal via stack
    while (snapshotStack.length) {
        /** @type {?} */
        const current = (/** @type {?} */ (snapshotStack.pop()));
        if (current === snapshot) {
            // we have found snapshot in the tree
            return true;
        }
        else {
            current.children.forEach((/**
             * @param {?} child
             * @return {?}
             */
            (child) => snapshotStack.push(child)));
        }
    }
    // we did not find snapshot in the tree
    return false;
}
/**
 * RouteTracker allows to
 * -- track changes of route query parameters;
 * -- update in-page data when route query parameters has changed;
 * -- update route query parameters when in-page data has been changed.
 *
 * <i>
 * Example:
 * Application has a page displaying list of entries filtered by parameters set on UI.
 * When one of filter parameters is changed => list of entries should be reloaded accordingly.
 * User can open one of entries in the separate page (leaving list of entries).
 * When he/she returns to the list (by pressing 'Back' browser's button)
 * it is convenient to have the latest parameters set on UI.
 *
 * Typical solution of this problem is to save all filters in query parameters of the page.
 * It allows to implement 'Back' behavior and gives 'bookmarkability' of such pages (at least).
 * </i>
 *
 * This class helps to solve similar problems.
 * In order to track and update query parameters RouteTracker registers {\@link RouteParamTracker}s.
 * Each {\@link RouteParamTracker} tracks and updates one or several query parameters.
 *
 * All registered {\@link RouteParamTracker}s are divided into filter and effect parameter trackers.
 *
 * When query parameters tracked by filter parameter trackers are changed => {\@link RouteTracker} emits the last filters
 * into {\@link RouteTracker#filterChanges} stream. This allows to update page based on filter changes.
 *
 * Effect parameter trackers also tracks query parameters, but they do not notify {\@link RouteTracker#filterChanges}
 * stream about the changes. This is the only difference.
 * @template F
 */
class RouteTracker {
    /**
     * @param {?} router
     * @param {?} filterTrackers
     * @param {?} effectTrackers
     * @param {?} defaultFilter
     */
    constructor(router, filterTrackers, effectTrackers, defaultFilter) {
        this.router = router;
        this.filterTrackers = filterTrackers;
        this.effectTrackers = effectTrackers;
        this.defaultFilter = defaultFilter;
    }
    /**
     * Returns true iff tracker listens for some ActivatedRoute
     * @return {?}
     */
    get listening() {
        return !isNil(this.subscription) && !this.subscription.closed;
    }
    /**
     * Returns the latest filters tracked by filter {\@link RouteParamTracker}s
     * @return {?}
     */
    get filterChanges() {
        if (!this.listening) {
            throw new Error('RouteTracker: filterChanges can be called only when tracker in listening state (listen(route))');
        }
        return this.filterChanges$$.asObservable();
    }
    /**
     * Constructs new {\@link RouteTracker} with the given set of {\@link RouteParamTracker}s as effect param trackers.
     * Set of filter {\@link RouteParamTracker}s is not modified.
     * @param {...?} effectTrackers
     * @return {?}
     */
    withEffects(...effectTrackers) {
        return new RouteTracker(this.router, this.filterTrackers, effectTrackers, this.defaultFilter);
    }
    /**
     * Starts listening for given {\@link ActivatedRoute}. Although this method returns {\@link Subscription}, generally
     * it is unnecessary to unsubscribe explicitly. When user leaves given route, subscription is destroyed automatically.
     * @param {?} route
     * @return {?}
     */
    listen(route) {
        if (this.listening) {
            throw new Error('RouteTracker: tracker already in listening state.');
        }
        const { filterTrackers, effectTrackers } = this;
        /** @type {?} */
        const subscription = this.subscription = new Subscription();
        this.filterChanges$$ = new BehaviorSubject(this.defaultFilter);
        this.enableMergeParams = true;
        // listen for route.queryParams
        // Generally speaking, it is unnecessarily to unsubscribe from route observables (as stated in the documentation
        // https://angular.io/guide/router#observable-parammap-and-component-reuse)
        subscription.add(route.queryParams.subscribe((/**
         * @param {?} queryParams
         * @return {?}
         */
        (queryParams) => {
            // merge queryParams into filter and effect trackers
            if (this.enableMergeParams) {
                filterTrackers.forEach((/**
                 * @param {?} tracker
                 * @return {?}
                 */
                (tracker) => tracker.mergeParams(queryParams)));
                effectTrackers.forEach((/**
                 * @param {?} tracker
                 * @return {?}
                 */
                (tracker) => tracker.mergeParams(queryParams)));
            }
        })));
        // Emit filter changes
        subscription.add(combineLatest(filterTrackers.map((/**
         * @param {?} tracker
         * @return {?}
         */
        (tracker) => tracker.filterChanges)))
            .subscribe((/**
         * @param {?} filters
         * @return {?}
         */
        (filters) => this.filterChanges$$.next((/** @type {?} */ (filters))))));
        // Listen for paramTrackers.paramChanges
        subscription.add(combineLatest(filterTrackers.concat(effectTrackers).map((/**
         * @param {?} tracker
         * @return {?}
         */
        (tracker) => tracker.paramChanges))).pipe(map((/**
         * @param {?} paramsArr
         * @return {?}
         */
        (paramsArr) => assign({}, ...paramsArr))), 
        // merge params into url
        map((/**
         * @param {?} params
         * @return {?}
         */
        (params) => this.mergeQueryParams(params))), 
        // if URL was modified
        filter(Boolean), 
        // there is no need to merge next query parameters, because they just changed
        tap((/**
         * @return {?}
         */
        () => {
            this.enableMergeParams = false;
        })), 
        // navigate
        switchMap((/**
         * @param {?} nextUrl
         * @return {?}
         */
        (nextUrl) => this.router.navigateByUrl(nextUrl, { replaceUrl: true }))))
            .subscribe((/**
         * @return {?}
         */
        () => {
            // enable parameter merging again
            this.enableMergeParams = true;
        })));
        // Create dispose subscription.
        // This subscription is necessary just to catch the moment when listened route was removed from the router state.
        // Unfortunately there is no easy way to catch the moment when ActivatedRoute should be disposed:
        // here we listen all NavigationEnd events and try to find if current route was removed from the router state.
        subscription.add(this.router.events.pipe(
        // wait when any navigation has been completed
        filter((/**
         * @param {?} event
         * @return {?}
         */
        (event) => event instanceof NavigationEnd)), 
        // if listened route was removed from router state => it means we left listened route
        filter((/**
         * @return {?}
         */
        () => !hasRouteInState(this.router.routerState, route.snapshot))))
            // dispose subscription if we leave listened route
            .subscribe((/**
         * @return {?}
         */
        () => subscription.unsubscribe())));
        return this.subscription;
    }
    /**
     * Dispose subscription for listened route.
     * @return {?}
     */
    unlisten() {
        if (this.listening) {
            this.filterChanges$$.complete();
            (/** @type {?} */ (this.subscription)).unsubscribe(); // tslint:disable-line:no-non-null-assertion
            this.subscription = void 0;
        }
    }
    /**
     * Merges given query parametes into current URL. If parameters do not change final URL this method returns undefined.
     * @private
     * @param {?} params
     * @return {?}
     */
    mergeQueryParams(params) {
        /** @type {?} */
        const url = this.router.routerState.snapshot.url;
        const { pathname, query, hash } = parseUrl(url);
        /** @type {?} */
        const nextQuery = Object.assign({}, query, params);
        /** @type {?} */
        const nextUrl = stringifyUrl({ pathname, query: nextQuery, hash });
        return nextUrl === url || isEqual(query, nextQuery) ? void 0 : nextUrl;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * This service allows to request and save dynamic data associated with routes.
 *
 * Service supports two kinds of data association:
 * - data can be associated with the route by its url and query parameters (current route)
 * - data can be associated with the state id attached to the current route as a parameter (savepoint)
 *
 */
class RouteDataService {
    /**
     * @param {?} router
     */
    constructor(router) {
        this.router = router;
    }
    /**
     * @param {...?} filters
     * @return {?}
     */
    tracker(...filters) {
        return new RouteTracker(this.router, filters, [], []);
    }
    /**
     * Merges provided <code>data</code> into the query parameters of current route
     * @template T
     * @param {?} data
     * @param {?=} mapper
     * @return {?}
     */
    mergeQueryParams(data, mapper = identityMapper()) {
        /** @type {?} */
        const url = this.router.routerState.snapshot.url;
        const { pathname, query, hash } = parseUrl(url);
        /** @type {?} */
        const serializedData = mapper.serialize(data);
        /** @type {?} */
        const nextUrl = stringifyUrl({ pathname, query: Object.assign({}, query, serializedData), hash });
        return nextUrl === url ? of(true) : from(this.router.navigateByUrl(nextUrl, {
            replaceUrl: true,
        }));
    }
}
RouteDataService.decorators = [
    { type: Injectable, args: [{ providedIn: 'root' },] }
];
/** @nocollapse */
RouteDataService.ctorParameters = () => [
    { type: Router }
];
/** @nocollapse */ RouteDataService.ngInjectableDef = ɵɵdefineInjectable({ factory: function RouteDataService_Factory() { return new RouteDataService(ɵɵinject(Router)); }, token: RouteDataService, providedIn: "root" });

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template F
 * @param {?} snapshot
 * @param {?=} paramNameOrDataMapper
 * @param {?=} optionalDataMapper
 * @return {?}
 */
function getRouteTrackerFilter(snapshot, paramNameOrDataMapper, optionalDataMapper) {
    /** @type {?} */
    const paramName = paramNameOrDataMapper && isString(paramNameOrDataMapper) ? paramNameOrDataMapper : 'filter';
    /** @type {?} */
    const dataMapper = (/** @type {?} */ ((optionalDataMapper || (paramName ? void 0 : paramNameOrDataMapper))));
    return deserializeData(snapshot.queryParams[paramName], dataMapper);
}
/**
 * @template F
 * @param {?} queryParamNameOrFilterable
 * @param {?=} filterableOrDataMapper
 * @param {?=} optionalDataMapper
 * @return {?}
 */
function filterTracker(queryParamNameOrFilterable, filterableOrDataMapper, optionalDataMapper) {
    if (!isString(queryParamNameOrFilterable)) {
        return filterTracker('filter', (/** @type {?} */ (queryParamNameOrFilterable)), (/** @type {?} */ (filterableOrDataMapper)));
    }
    /** @type {?} */
    const queryParamName = (/** @type {?} */ (queryParamNameOrFilterable));
    /** @type {?} */
    const filterable = (/** @type {?} */ (filterableOrDataMapper));
    /** @type {?} */
    const dataMapper = (/** @type {?} */ (optionalDataMapper)) || identityMapper();
    /** @type {?} */
    let defaultFilter = filterable.currentFilter;
    /** @type {?} */
    const tracker = {
        filterChanges: filterable.filterChanges,
        paramChanges: filterable.filterChanges.pipe(
        // there is no sense to store initial filter in query parameters
        skip(1), map((/**
         * @param {?} filter
         * @return {?}
         */
        (filter) => isSimilarFilter(filter, defaultFilter)
            // send empty set of query parameters if they are equal to default filter
            ? { [queryParamName]: void 0 }
            // serialize filter
            // store whole filter under 'filter' query parameter
            : ({ [queryParamName]: dataMapper.serialize(filter) }))), startWith({})),
        // take parameters from 'filter' query parameter
        mergeParams: (/**
         * @param {?} params
         * @return {?}
         */
        (params) => {
            /** @type {?} */
            const filter = params[queryParamName];
            if (filter) {
                filterable.setFilter(dataMapper.deserialize(filter));
            }
            else if (!isNil(defaultFilter)) {
                filterable.setFilter(defaultFilter);
            }
        }),
        withDefault: (/**
         * @param {?} filter
         * @return {?}
         */
        (filter) => {
            defaultFilter = filter;
            return tracker;
        }),
    };
    return tracker;
}
/**
 * @template F
 * @param {?} queryParamNameOrControl
 * @param {?=} controlOrDataMapper
 * @param {?=} optionalDataMapper
 * @return {?}
 */
function controlTracker(queryParamNameOrControl, controlOrDataMapper, optionalDataMapper) {
    if (!isString(queryParamNameOrControl)) {
        return controlTracker('filter', (/** @type {?} */ (queryParamNameOrControl)), (/** @type {?} */ (controlOrDataMapper)));
    }
    /** @type {?} */
    const queryParamName = (/** @type {?} */ (queryParamNameOrControl));
    /** @type {?} */
    const control = (/** @type {?} */ (controlOrDataMapper));
    /** @type {?} */
    const dataMapper = (/** @type {?} */ (optionalDataMapper)) || identityMapper();
    /** @type {?} */
    let defaultFilter = control.value;
    // we have to make lazy stream from FormControl in order to get actual current value on the moment of subscription
    /** @type {?} */
    const value$ = of(control).pipe(switchMap((/**
     * @param {?} c
     * @return {?}
     */
    (c) => c.valueChanges.pipe(startWith(c.value)))));
    /** @type {?} */
    const tracker = {
        filterChanges: value$,
        paramChanges: value$.pipe(
        // there is no sense to store initial filter in query parameters
        skip(1), map((/**
         * @param {?} filter
         * @return {?}
         */
        (filter) => isSimilarFilter(filter, defaultFilter)
            // send empty set of query parameters if they are equal to default filter
            ? { [queryParamName]: void 0 }
            // serialize filter
            // store whole filter under 'filter' query parameter
            : ({ [queryParamName]: dataMapper.serialize(filter) }))), 
        // start with empty set of query parameters
        startWith({})),
        // take parameters from 'filter' query parameter
        mergeParams: (/**
         * @param {?} params
         * @return {?}
         */
        (params) => {
            /** @type {?} */
            const filter = params[queryParamName];
            if (filter) {
                control.setValue(dataMapper.deserialize(filter));
            }
            else if (!isNil(defaultFilter)) {
                control.setValue(defaultFilter);
            }
        }),
        withDefault: (/**
         * @param {?} filter
         * @return {?}
         */
        (filter) => {
            defaultFilter = filter;
            return tracker;
        }),
    };
    return tracker;
}
/**
 * @template F
 * @param {?} control
 * @return {?}
 */
function filterableControl(control) {
    /** @type {?} */
    const filterChanges$$ = new BehaviorSubject(control.value);
    /**
     * @param {?} filter
     * @return {?}
     */
    function setFilter(filter) {
        delegated.patchValue(filter);
        emitFilter();
    }
    /**
     * @return {?}
     */
    function emitFilter() {
        filterChanges$$.next(delegated.value);
    }
    /** @type {?} */
    const delegated = create(control, {
        /**
         * @return {?}
         */
        get currentFilter() {
            return control.value;
        },
        filterChanges: filterChanges$$.asObservable(),
        setFilter,
        emitFilter,
    });
    return delegated;
}
/**
 * @template F
 * @param {?} queryParamNameOrDataMapper
 * @param {?=} optionalDataMapper
 * @return {?}
 */
function manualTracker(queryParamNameOrDataMapper, optionalDataMapper) {
    if (!isString(queryParamNameOrDataMapper)) {
        return manualTracker('filter', (/** @type {?} */ (queryParamNameOrDataMapper)));
    }
    /** @type {?} */
    const queryParamName = (/** @type {?} */ (queryParamNameOrDataMapper));
    /** @type {?} */
    const dataMapper = (/** @type {?} */ (optionalDataMapper)) || identityMapper();
    /** @type {?} */
    const control = new FormControl();
    return create(controlTracker(queryParamName, control, dataMapper), {
        setValue: (/**
         * @param {?} filter
         * @return {?}
         */
        (filter) => control.setValue(filter)),
    });
}
/**
 * @template F
 * @param {?} filterA
 * @param {?} filterB
 * @return {?}
 */
function isSimilarFilter(filterA, filterB) {
    return isEqual(filterA, filterB);
}
/**
 * Defines {\@link RouteParamTracker} which updates query parameters from the given {\@link Observable}.
 * If query parameters is changed tracker calls {\@link ObservableFilter#setFilter} method
 * to allow developer to update a page stage.
 * Filter coming from/to query parameters will be deserialized/serialized using provided <code>dataMapper</code>.
 * This method stores filter under the single query parameter (by default, 'filter')
 * @template F
 * @param {?} oFilter
 * @return {?}
 */
function observableFilterTracker(oFilter) {
    /** @type {?} */
    const queryParamName = oFilter.name || 'filter';
    /** @type {?} */
    const dataMapper = oFilter.dataMapper || identityMapper();
    /** @type {?} */
    let defaultFilter = (/** @type {?} */ (void 0));
    /** @type {?} */
    const value$ = oFilter.filterChanges;
    /** @type {?} */
    const tracker = {
        filterChanges: value$,
        paramChanges: value$.pipe(
        // there is no sense to store initial filter in query parameters
        skip(1), map((/**
         * @param {?} filter
         * @return {?}
         */
        (filter) => isSimilarFilter(filter, defaultFilter)
            // send empty set of query parameters if they are equal to default filter
            ? { [queryParamName]: void 0 }
            // serialize filter
            // store whole filter under 'filter' query parameter
            : ({ [queryParamName]: dataMapper.serialize(filter) }))), 
        // start with empty set of query parameters
        startWith({})),
        // take parameters from 'filter' query parameter
        mergeParams: (/**
         * @param {?} params
         * @return {?}
         */
        (params) => {
            const { setFilter } = oFilter;
            if (setFilter) {
                /** @type {?} */
                const filter = params[queryParamName];
                if (filter) {
                    setFilter(dataMapper.deserialize(filter));
                }
                else if (!isNil(defaultFilter)) {
                    setFilter(defaultFilter);
                }
            }
        }),
        withDefault: (/**
         * @param {?} filter
         * @return {?}
         */
        (filter) => {
            defaultFilter = filter;
            return tracker;
        }),
    };
    return tracker;
}
/**
 * Defines {\@link RouteParamTracker} which updates query parameters from the given {\@link Observable}.
 * If query parameters is changed tracker calls {\@link ObservableFilter#setFilter} method
 * to allow developer to update a page stage.
 * Filter coming from/to query parameters will be deserialized/serialized using provided <code>dataMappers</code>.
 * This tracker tracks only parameters specified under <code>names</code> parameter.
 * Modified filter is merged into query parameters defined by <code>names</code> parameter.
 * @template F
 * @param {?} oFilter
 * @return {?}
 */
function observablePlainFilterTracker(oFilter) {
    /** @type {?} */
    let defaultFilter = (/** @type {?} */ (void 0));
    const { dataMappers } = oFilter;
    /** @type {?} */
    const serialize = (/**
     * @param {?} name
     * @param {?} value
     * @return {?}
     */
    (name, value) => {
        return serializeData(value, dataMappers && dataMappers[name]);
    });
    /** @type {?} */
    const deserialize = (/**
     * @param {?} name
     * @param {?} value
     * @return {?}
     */
    (name, value) => {
        return deserializeData(value, dataMappers && dataMappers[name]);
    });
    /** @type {?} */
    const emptyQueryParameters = fromPairs(oFilter.names.map((/**
     * @param {?} name
     * @return {?}
     */
    (name) => [name, void 0])));
    /** @type {?} */
    const value$ = oFilter.filterChanges;
    /** @type {?} */
    const tracker = {
        filterChanges: value$,
        paramChanges: value$.pipe(
        // there is no sense to store initial filter in query parameters
        skip(1), map((/**
         * @param {?} filter
         * @return {?}
         */
        (filter) => isSimilarFilter(filter, defaultFilter)
            // send empty set of query parameters if they are equal to default filter
            ? emptyQueryParameters
            // serialize filter
            // store whole filter under corresponding query parameters
            : fromPairs(oFilter.names
                .map((/**
             * @param {?} name
             * @return {?}
             */
            (name) => [name, serialize(name, filter[name])]))
                // send empty query parameter if it is equal to corresponding parameter from the default filter
                .map((/**
             * @param {?} __0
             * @return {?}
             */
            ([name, value]) => [
                name,
                defaultFilter && isSimilarFilter(value, defaultFilter[name]) ? void 0 : value,
            ]))))), 
        // start with empty set of query parameters
        startWith({})),
        // take parameters from corresponding query parameters
        mergeParams: (/**
         * @param {?} params
         * @return {?}
         */
        (params) => {
            const { setFilter } = oFilter;
            if (setFilter) {
                /** @type {?} */
                const filter = fromPairs(oFilter.names.map((/**
                 * @param {?} name
                 * @return {?}
                 */
                (name) => [name, deserialize(name, params[name])])));
                if (!isEmpty(filter) && !every(filter, isNil)) {
                    setFilter((/** @type {?} */ (filter)));
                }
                else if (!isNil(defaultFilter)) {
                    setFilter(defaultFilter);
                }
            }
        }),
        withDefault: (/**
         * @param {?} filter
         * @return {?}
         */
        (filter) => {
            defaultFilter = filter;
            return tracker;
        }),
    };
    return tracker;
}

export { RouteDataService, RouteTracker, controlTracker, filterTracker, filterableControl, getRouteTrackerFilter, manualTracker, observableFilterTracker, observablePlainFilterTracker };
//# sourceMappingURL=softeq-angular-route-data.js.map
