import { __read, __spread, __assign } from 'tslib';
import { Injectable, ɵɵdefineInjectable, ɵɵinject } from '@angular/core';
import { NavigationEnd, Router } from '@angular/router';
import { identityMapper, deserializeData, serializeData } from '@softeq/data-mappers';
import { Subscription, BehaviorSubject, combineLatest, of, from } from 'rxjs';
import assign from 'lodash/assign';
import isEqual from 'lodash/isEqual';
import isNil from 'lodash/isNil';
import { map, filter, tap, switchMap, skip, startWith } from 'rxjs/operators';
import flatMap from 'lodash/flatMap';
import groupBy from 'lodash/groupBy';
import mapValues from 'lodash/mapValues';
import toPairs from 'lodash/toPairs';
import { FormControl } from '@angular/forms';
import create from 'lodash/create';
import every from 'lodash/every';
import fromPairs from 'lodash/fromPairs';
import isEmpty from 'lodash/isEmpty';
import isString from 'lodash/isString';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
var URL_RE = /^([^?#]*)(?:\?([^#]*))?(?:#(.*))?$/;
/**
 * @param {?} url
 * @return {?}
 */
function parseUrl(url) {
    /** @type {?} */
    var match = URL_RE.exec(url);
    if (match) {
        var _a = __read(match, 4), _1 = _a[0], pathname = _a[1], query = _a[2], hash = _a[3];
        return {
            pathname: pathname,
            hash: hash,
            query: mapValues(mapValues(groupBy((query || '')
                .split('&')
                .map((/**
             * @param {?} param
             * @return {?}
             */
            function (param) { return param.split('=').map(decodeURIComponent); })), (/**
             * @param {?} __0
             * @return {?}
             */
            function (_a) {
                var _b = __read(_a, 1), key = _b[0];
                return key;
            })), (/**
             * @param {?} pairs
             * @return {?}
             */
            function (pairs) { return pairs.map((/**
             * @param {?} __0
             * @return {?}
             */
            function (_a) {
                var _b = __read(_a, 2), _ = _b[0], value = _b[1];
                return value;
            })); })), (/**
             * @param {?} values
             * @return {?}
             */
            function (values) { return values.length === 1 ? values[0] : values; })),
        };
    }
    else {
        return { pathname: '', query: {}, hash: '' };
    }
}
/**
 * @param {?} url
 * @return {?}
 */
function stringifyUrl(url) {
    return [
        url.pathname,
        url.query ? '?' : '',
        flatMap(toPairs(url.query)
            .filter((/**
         * @param {?} __0
         * @return {?}
         */
        function (_a) {
            var _b = __read(_a, 2), _ = _b[0], value = _b[1];
            return !isNil(value);
        }))
            .map((/**
         * @param {?} __0
         * @return {?}
         */
        function (_a) {
            var _b = __read(_a, 2), key = _b[0], value = _b[1];
            return [key, Array.isArray(value) ? value : [value]];
        })), (/**
         * @param {?} __0
         * @return {?}
         */
        function (_a) {
            var _b = __read(_a, 2), key = _b[0], values = _b[1];
            return values.map((/**
             * @param {?} value
             * @return {?}
             */
            function (value) { return encodeURIComponent(key) + "=" + encodeURIComponent((value)); }));
        })).join('&'),
        url.hash ? '#' : '',
        url.hash,
    ].join('');
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Checks if provided snapshot is in the snapshot tree of the router state
 * @param {?} router
 * @param {?} snapshot
 * @return {?}
 */
function hasRouteInState(router, snapshot) {
    /** @type {?} */
    var snapshotStack = [];
    snapshotStack.push(router.snapshot.root);
    // simple tree traversal via stack
    while (snapshotStack.length) {
        /** @type {?} */
        var current = (/** @type {?} */ (snapshotStack.pop()));
        if (current === snapshot) {
            // we have found snapshot in the tree
            return true;
        }
        else {
            current.children.forEach((/**
             * @param {?} child
             * @return {?}
             */
            function (child) { return snapshotStack.push(child); }));
        }
    }
    // we did not find snapshot in the tree
    return false;
}
/**
 * RouteTracker allows to
 * -- track changes of route query parameters;
 * -- update in-page data when route query parameters has changed;
 * -- update route query parameters when in-page data has been changed.
 *
 * <i>
 * Example:
 * Application has a page displaying list of entries filtered by parameters set on UI.
 * When one of filter parameters is changed => list of entries should be reloaded accordingly.
 * User can open one of entries in the separate page (leaving list of entries).
 * When he/she returns to the list (by pressing 'Back' browser's button)
 * it is convenient to have the latest parameters set on UI.
 *
 * Typical solution of this problem is to save all filters in query parameters of the page.
 * It allows to implement 'Back' behavior and gives 'bookmarkability' of such pages (at least).
 * </i>
 *
 * This class helps to solve similar problems.
 * In order to track and update query parameters RouteTracker registers {\@link RouteParamTracker}s.
 * Each {\@link RouteParamTracker} tracks and updates one or several query parameters.
 *
 * All registered {\@link RouteParamTracker}s are divided into filter and effect parameter trackers.
 *
 * When query parameters tracked by filter parameter trackers are changed => {\@link RouteTracker} emits the last filters
 * into {\@link RouteTracker#filterChanges} stream. This allows to update page based on filter changes.
 *
 * Effect parameter trackers also tracks query parameters, but they do not notify {\@link RouteTracker#filterChanges}
 * stream about the changes. This is the only difference.
 * @template F
 */
var  /**
 * RouteTracker allows to
 * -- track changes of route query parameters;
 * -- update in-page data when route query parameters has changed;
 * -- update route query parameters when in-page data has been changed.
 *
 * <i>
 * Example:
 * Application has a page displaying list of entries filtered by parameters set on UI.
 * When one of filter parameters is changed => list of entries should be reloaded accordingly.
 * User can open one of entries in the separate page (leaving list of entries).
 * When he/she returns to the list (by pressing 'Back' browser's button)
 * it is convenient to have the latest parameters set on UI.
 *
 * Typical solution of this problem is to save all filters in query parameters of the page.
 * It allows to implement 'Back' behavior and gives 'bookmarkability' of such pages (at least).
 * </i>
 *
 * This class helps to solve similar problems.
 * In order to track and update query parameters RouteTracker registers {\@link RouteParamTracker}s.
 * Each {\@link RouteParamTracker} tracks and updates one or several query parameters.
 *
 * All registered {\@link RouteParamTracker}s are divided into filter and effect parameter trackers.
 *
 * When query parameters tracked by filter parameter trackers are changed => {\@link RouteTracker} emits the last filters
 * into {\@link RouteTracker#filterChanges} stream. This allows to update page based on filter changes.
 *
 * Effect parameter trackers also tracks query parameters, but they do not notify {\@link RouteTracker#filterChanges}
 * stream about the changes. This is the only difference.
 * @template F
 */
RouteTracker = /** @class */ (function () {
    function RouteTracker(router, filterTrackers, effectTrackers, defaultFilter) {
        this.router = router;
        this.filterTrackers = filterTrackers;
        this.effectTrackers = effectTrackers;
        this.defaultFilter = defaultFilter;
    }
    Object.defineProperty(RouteTracker.prototype, "listening", {
        /**
         * Returns true iff tracker listens for some ActivatedRoute
         */
        get: /**
         * Returns true iff tracker listens for some ActivatedRoute
         * @return {?}
         */
        function () {
            return !isNil(this.subscription) && !this.subscription.closed;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(RouteTracker.prototype, "filterChanges", {
        /**
         * Returns the latest filters tracked by filter {@link RouteParamTracker}s
         */
        get: /**
         * Returns the latest filters tracked by filter {\@link RouteParamTracker}s
         * @return {?}
         */
        function () {
            if (!this.listening) {
                throw new Error('RouteTracker: filterChanges can be called only when tracker in listening state (listen(route))');
            }
            return this.filterChanges$$.asObservable();
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Constructs new {@link RouteTracker} with the given set of {@link RouteParamTracker}s as effect param trackers.
     * Set of filter {@link RouteParamTracker}s is not modified.
     */
    /**
     * Constructs new {\@link RouteTracker} with the given set of {\@link RouteParamTracker}s as effect param trackers.
     * Set of filter {\@link RouteParamTracker}s is not modified.
     * @param {...?} effectTrackers
     * @return {?}
     */
    RouteTracker.prototype.withEffects = /**
     * Constructs new {\@link RouteTracker} with the given set of {\@link RouteParamTracker}s as effect param trackers.
     * Set of filter {\@link RouteParamTracker}s is not modified.
     * @param {...?} effectTrackers
     * @return {?}
     */
    function () {
        var effectTrackers = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            effectTrackers[_i] = arguments[_i];
        }
        return new RouteTracker(this.router, this.filterTrackers, effectTrackers, this.defaultFilter);
    };
    /**
     * Starts listening for given {@link ActivatedRoute}. Although this method returns {@link Subscription}, generally
     * it is unnecessary to unsubscribe explicitly. When user leaves given route, subscription is destroyed automatically.
     */
    /**
     * Starts listening for given {\@link ActivatedRoute}. Although this method returns {\@link Subscription}, generally
     * it is unnecessary to unsubscribe explicitly. When user leaves given route, subscription is destroyed automatically.
     * @param {?} route
     * @return {?}
     */
    RouteTracker.prototype.listen = /**
     * Starts listening for given {\@link ActivatedRoute}. Although this method returns {\@link Subscription}, generally
     * it is unnecessary to unsubscribe explicitly. When user leaves given route, subscription is destroyed automatically.
     * @param {?} route
     * @return {?}
     */
    function (route) {
        var _this = this;
        if (this.listening) {
            throw new Error('RouteTracker: tracker already in listening state.');
        }
        var _a = this, filterTrackers = _a.filterTrackers, effectTrackers = _a.effectTrackers;
        /** @type {?} */
        var subscription = this.subscription = new Subscription();
        this.filterChanges$$ = new BehaviorSubject(this.defaultFilter);
        this.enableMergeParams = true;
        // listen for route.queryParams
        // Generally speaking, it is unnecessarily to unsubscribe from route observables (as stated in the documentation
        // https://angular.io/guide/router#observable-parammap-and-component-reuse)
        subscription.add(route.queryParams.subscribe((/**
         * @param {?} queryParams
         * @return {?}
         */
        function (queryParams) {
            // merge queryParams into filter and effect trackers
            if (_this.enableMergeParams) {
                filterTrackers.forEach((/**
                 * @param {?} tracker
                 * @return {?}
                 */
                function (tracker) { return tracker.mergeParams(queryParams); }));
                effectTrackers.forEach((/**
                 * @param {?} tracker
                 * @return {?}
                 */
                function (tracker) { return tracker.mergeParams(queryParams); }));
            }
        })));
        // Emit filter changes
        subscription.add(combineLatest(filterTrackers.map((/**
         * @param {?} tracker
         * @return {?}
         */
        function (tracker) { return tracker.filterChanges; })))
            .subscribe((/**
         * @param {?} filters
         * @return {?}
         */
        function (filters) { return _this.filterChanges$$.next((/** @type {?} */ (filters))); })));
        // Listen for paramTrackers.paramChanges
        subscription.add(combineLatest(filterTrackers.concat(effectTrackers).map((/**
         * @param {?} tracker
         * @return {?}
         */
        function (tracker) { return tracker.paramChanges; }))).pipe(map((/**
         * @param {?} paramsArr
         * @return {?}
         */
        function (paramsArr) { return assign.apply(void 0, __spread([{}], paramsArr)); })), 
        // merge params into url
        map((/**
         * @param {?} params
         * @return {?}
         */
        function (params) { return _this.mergeQueryParams(params); })), 
        // if URL was modified
        filter(Boolean), 
        // there is no need to merge next query parameters, because they just changed
        tap((/**
         * @return {?}
         */
        function () {
            _this.enableMergeParams = false;
        })), 
        // navigate
        switchMap((/**
         * @param {?} nextUrl
         * @return {?}
         */
        function (nextUrl) { return _this.router.navigateByUrl(nextUrl, { replaceUrl: true }); })))
            .subscribe((/**
         * @return {?}
         */
        function () {
            // enable parameter merging again
            _this.enableMergeParams = true;
        })));
        // Create dispose subscription.
        // This subscription is necessary just to catch the moment when listened route was removed from the router state.
        // Unfortunately there is no easy way to catch the moment when ActivatedRoute should be disposed:
        // here we listen all NavigationEnd events and try to find if current route was removed from the router state.
        subscription.add(this.router.events.pipe(
        // wait when any navigation has been completed
        filter((/**
         * @param {?} event
         * @return {?}
         */
        function (event) { return event instanceof NavigationEnd; })), 
        // if listened route was removed from router state => it means we left listened route
        filter((/**
         * @return {?}
         */
        function () { return !hasRouteInState(_this.router.routerState, route.snapshot); })))
            // dispose subscription if we leave listened route
            .subscribe((/**
         * @return {?}
         */
        function () { return subscription.unsubscribe(); })));
        return this.subscription;
    };
    /**
     * Dispose subscription for listened route.
     */
    /**
     * Dispose subscription for listened route.
     * @return {?}
     */
    RouteTracker.prototype.unlisten = /**
     * Dispose subscription for listened route.
     * @return {?}
     */
    function () {
        if (this.listening) {
            this.filterChanges$$.complete();
            (/** @type {?} */ (this.subscription)).unsubscribe(); // tslint:disable-line:no-non-null-assertion
            this.subscription = void 0;
        }
    };
    /**
     * Merges given query parametes into current URL. If parameters do not change final URL this method returns undefined.
     */
    /**
     * Merges given query parametes into current URL. If parameters do not change final URL this method returns undefined.
     * @private
     * @param {?} params
     * @return {?}
     */
    RouteTracker.prototype.mergeQueryParams = /**
     * Merges given query parametes into current URL. If parameters do not change final URL this method returns undefined.
     * @private
     * @param {?} params
     * @return {?}
     */
    function (params) {
        /** @type {?} */
        var url = this.router.routerState.snapshot.url;
        var _a = parseUrl(url), pathname = _a.pathname, query = _a.query, hash = _a.hash;
        /** @type {?} */
        var nextQuery = __assign({}, query, params);
        /** @type {?} */
        var nextUrl = stringifyUrl({ pathname: pathname, query: nextQuery, hash: hash });
        return nextUrl === url || isEqual(query, nextQuery) ? void 0 : nextUrl;
    };
    return RouteTracker;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * This service allows to request and save dynamic data associated with routes.
 *
 * Service supports two kinds of data association:
 * - data can be associated with the route by its url and query parameters (current route)
 * - data can be associated with the state id attached to the current route as a parameter (savepoint)
 *
 */
var RouteDataService = /** @class */ (function () {
    function RouteDataService(router) {
        this.router = router;
    }
    /**
     * @param {...?} filters
     * @return {?}
     */
    RouteDataService.prototype.tracker = /**
     * @param {...?} filters
     * @return {?}
     */
    function () {
        var filters = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            filters[_i] = arguments[_i];
        }
        return new RouteTracker(this.router, filters, [], []);
    };
    /**
     * Merges provided <code>data</code> into the query parameters of current route
     */
    /**
     * Merges provided <code>data</code> into the query parameters of current route
     * @template T
     * @param {?} data
     * @param {?=} mapper
     * @return {?}
     */
    RouteDataService.prototype.mergeQueryParams = /**
     * Merges provided <code>data</code> into the query parameters of current route
     * @template T
     * @param {?} data
     * @param {?=} mapper
     * @return {?}
     */
    function (data, mapper) {
        if (mapper === void 0) { mapper = identityMapper(); }
        /** @type {?} */
        var url = this.router.routerState.snapshot.url;
        var _a = parseUrl(url), pathname = _a.pathname, query = _a.query, hash = _a.hash;
        /** @type {?} */
        var serializedData = mapper.serialize(data);
        /** @type {?} */
        var nextUrl = stringifyUrl({ pathname: pathname, query: __assign({}, query, serializedData), hash: hash });
        return nextUrl === url ? of(true) : from(this.router.navigateByUrl(nextUrl, {
            replaceUrl: true,
        }));
    };
    RouteDataService.decorators = [
        { type: Injectable, args: [{ providedIn: 'root' },] }
    ];
    /** @nocollapse */
    RouteDataService.ctorParameters = function () { return [
        { type: Router }
    ]; };
    /** @nocollapse */ RouteDataService.ngInjectableDef = ɵɵdefineInjectable({ factory: function RouteDataService_Factory() { return new RouteDataService(ɵɵinject(Router)); }, token: RouteDataService, providedIn: "root" });
    return RouteDataService;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template F
 * @param {?} snapshot
 * @param {?=} paramNameOrDataMapper
 * @param {?=} optionalDataMapper
 * @return {?}
 */
function getRouteTrackerFilter(snapshot, paramNameOrDataMapper, optionalDataMapper) {
    /** @type {?} */
    var paramName = paramNameOrDataMapper && isString(paramNameOrDataMapper) ? paramNameOrDataMapper : 'filter';
    /** @type {?} */
    var dataMapper = (/** @type {?} */ ((optionalDataMapper || (paramName ? void 0 : paramNameOrDataMapper))));
    return deserializeData(snapshot.queryParams[paramName], dataMapper);
}
/**
 * @template F
 * @param {?} queryParamNameOrFilterable
 * @param {?=} filterableOrDataMapper
 * @param {?=} optionalDataMapper
 * @return {?}
 */
function filterTracker(queryParamNameOrFilterable, filterableOrDataMapper, optionalDataMapper) {
    if (!isString(queryParamNameOrFilterable)) {
        return filterTracker('filter', (/** @type {?} */ (queryParamNameOrFilterable)), (/** @type {?} */ (filterableOrDataMapper)));
    }
    /** @type {?} */
    var queryParamName = (/** @type {?} */ (queryParamNameOrFilterable));
    /** @type {?} */
    var filterable = (/** @type {?} */ (filterableOrDataMapper));
    /** @type {?} */
    var dataMapper = (/** @type {?} */ (optionalDataMapper)) || identityMapper();
    /** @type {?} */
    var defaultFilter = filterable.currentFilter;
    /** @type {?} */
    var tracker = {
        filterChanges: filterable.filterChanges,
        paramChanges: filterable.filterChanges.pipe(
        // there is no sense to store initial filter in query parameters
        skip(1), map((/**
         * @param {?} filter
         * @return {?}
         */
        function (filter) {
            var _a, _b;
            return isSimilarFilter(filter, defaultFilter)
                // send empty set of query parameters if they are equal to default filter
                ? (_a = {}, _a[queryParamName] = void 0, _a) 
            // serialize filter
            // store whole filter under 'filter' query parameter
            : (_b = {}, _b[queryParamName] = dataMapper.serialize(filter), _b);
        })), startWith({})),
        // take parameters from 'filter' query parameter
        mergeParams: (/**
         * @param {?} params
         * @return {?}
         */
        function (params) {
            /** @type {?} */
            var filter = params[queryParamName];
            if (filter) {
                filterable.setFilter(dataMapper.deserialize(filter));
            }
            else if (!isNil(defaultFilter)) {
                filterable.setFilter(defaultFilter);
            }
        }),
        withDefault: (/**
         * @param {?} filter
         * @return {?}
         */
        function (filter) {
            defaultFilter = filter;
            return tracker;
        }),
    };
    return tracker;
}
/**
 * @template F
 * @param {?} queryParamNameOrControl
 * @param {?=} controlOrDataMapper
 * @param {?=} optionalDataMapper
 * @return {?}
 */
function controlTracker(queryParamNameOrControl, controlOrDataMapper, optionalDataMapper) {
    if (!isString(queryParamNameOrControl)) {
        return controlTracker('filter', (/** @type {?} */ (queryParamNameOrControl)), (/** @type {?} */ (controlOrDataMapper)));
    }
    /** @type {?} */
    var queryParamName = (/** @type {?} */ (queryParamNameOrControl));
    /** @type {?} */
    var control = (/** @type {?} */ (controlOrDataMapper));
    /** @type {?} */
    var dataMapper = (/** @type {?} */ (optionalDataMapper)) || identityMapper();
    /** @type {?} */
    var defaultFilter = control.value;
    // we have to make lazy stream from FormControl in order to get actual current value on the moment of subscription
    /** @type {?} */
    var value$ = of(control).pipe(switchMap((/**
     * @param {?} c
     * @return {?}
     */
    function (c) { return c.valueChanges.pipe(startWith(c.value)); })));
    /** @type {?} */
    var tracker = {
        filterChanges: value$,
        paramChanges: value$.pipe(
        // there is no sense to store initial filter in query parameters
        skip(1), map((/**
         * @param {?} filter
         * @return {?}
         */
        function (filter) {
            var _a, _b;
            return isSimilarFilter(filter, defaultFilter)
                // send empty set of query parameters if they are equal to default filter
                ? (_a = {}, _a[queryParamName] = void 0, _a) 
            // serialize filter
            // store whole filter under 'filter' query parameter
            : (_b = {}, _b[queryParamName] = dataMapper.serialize(filter), _b);
        })), 
        // start with empty set of query parameters
        startWith({})),
        // take parameters from 'filter' query parameter
        mergeParams: (/**
         * @param {?} params
         * @return {?}
         */
        function (params) {
            /** @type {?} */
            var filter = params[queryParamName];
            if (filter) {
                control.setValue(dataMapper.deserialize(filter));
            }
            else if (!isNil(defaultFilter)) {
                control.setValue(defaultFilter);
            }
        }),
        withDefault: (/**
         * @param {?} filter
         * @return {?}
         */
        function (filter) {
            defaultFilter = filter;
            return tracker;
        }),
    };
    return tracker;
}
/**
 * @template F
 * @param {?} control
 * @return {?}
 */
function filterableControl(control) {
    /** @type {?} */
    var filterChanges$$ = new BehaviorSubject(control.value);
    /**
     * @param {?} filter
     * @return {?}
     */
    function setFilter(filter) {
        delegated.patchValue(filter);
        emitFilter();
    }
    /**
     * @return {?}
     */
    function emitFilter() {
        filterChanges$$.next(delegated.value);
    }
    /** @type {?} */
    var delegated = create(control, {
        /**
         * @return {?}
         */
        get currentFilter() {
            return control.value;
        },
        filterChanges: filterChanges$$.asObservable(),
        setFilter: setFilter,
        emitFilter: emitFilter,
    });
    return delegated;
}
/**
 * @template F
 * @param {?} queryParamNameOrDataMapper
 * @param {?=} optionalDataMapper
 * @return {?}
 */
function manualTracker(queryParamNameOrDataMapper, optionalDataMapper) {
    if (!isString(queryParamNameOrDataMapper)) {
        return manualTracker('filter', (/** @type {?} */ (queryParamNameOrDataMapper)));
    }
    /** @type {?} */
    var queryParamName = (/** @type {?} */ (queryParamNameOrDataMapper));
    /** @type {?} */
    var dataMapper = (/** @type {?} */ (optionalDataMapper)) || identityMapper();
    /** @type {?} */
    var control = new FormControl();
    return create(controlTracker(queryParamName, control, dataMapper), {
        setValue: (/**
         * @param {?} filter
         * @return {?}
         */
        function (filter) { return control.setValue(filter); }),
    });
}
/**
 * @template F
 * @param {?} filterA
 * @param {?} filterB
 * @return {?}
 */
function isSimilarFilter(filterA, filterB) {
    return isEqual(filterA, filterB);
}
/**
 * Defines {\@link RouteParamTracker} which updates query parameters from the given {\@link Observable}.
 * If query parameters is changed tracker calls {\@link ObservableFilter#setFilter} method
 * to allow developer to update a page stage.
 * Filter coming from/to query parameters will be deserialized/serialized using provided <code>dataMapper</code>.
 * This method stores filter under the single query parameter (by default, 'filter')
 * @template F
 * @param {?} oFilter
 * @return {?}
 */
function observableFilterTracker(oFilter) {
    /** @type {?} */
    var queryParamName = oFilter.name || 'filter';
    /** @type {?} */
    var dataMapper = oFilter.dataMapper || identityMapper();
    /** @type {?} */
    var defaultFilter = (/** @type {?} */ (void 0));
    /** @type {?} */
    var value$ = oFilter.filterChanges;
    /** @type {?} */
    var tracker = {
        filterChanges: value$,
        paramChanges: value$.pipe(
        // there is no sense to store initial filter in query parameters
        skip(1), map((/**
         * @param {?} filter
         * @return {?}
         */
        function (filter) {
            var _a, _b;
            return isSimilarFilter(filter, defaultFilter)
                // send empty set of query parameters if they are equal to default filter
                ? (_a = {}, _a[queryParamName] = void 0, _a) 
            // serialize filter
            // store whole filter under 'filter' query parameter
            : (_b = {}, _b[queryParamName] = dataMapper.serialize(filter), _b);
        })), 
        // start with empty set of query parameters
        startWith({})),
        // take parameters from 'filter' query parameter
        mergeParams: (/**
         * @param {?} params
         * @return {?}
         */
        function (params) {
            var setFilter = oFilter.setFilter;
            if (setFilter) {
                /** @type {?} */
                var filter = params[queryParamName];
                if (filter) {
                    setFilter(dataMapper.deserialize(filter));
                }
                else if (!isNil(defaultFilter)) {
                    setFilter(defaultFilter);
                }
            }
        }),
        withDefault: (/**
         * @param {?} filter
         * @return {?}
         */
        function (filter) {
            defaultFilter = filter;
            return tracker;
        }),
    };
    return tracker;
}
/**
 * Defines {\@link RouteParamTracker} which updates query parameters from the given {\@link Observable}.
 * If query parameters is changed tracker calls {\@link ObservableFilter#setFilter} method
 * to allow developer to update a page stage.
 * Filter coming from/to query parameters will be deserialized/serialized using provided <code>dataMappers</code>.
 * This tracker tracks only parameters specified under <code>names</code> parameter.
 * Modified filter is merged into query parameters defined by <code>names</code> parameter.
 * @template F
 * @param {?} oFilter
 * @return {?}
 */
function observablePlainFilterTracker(oFilter) {
    /** @type {?} */
    var defaultFilter = (/** @type {?} */ (void 0));
    var dataMappers = oFilter.dataMappers;
    /** @type {?} */
    var serialize = (/**
     * @param {?} name
     * @param {?} value
     * @return {?}
     */
    function (name, value) {
        return serializeData(value, dataMappers && dataMappers[name]);
    });
    /** @type {?} */
    var deserialize = (/**
     * @param {?} name
     * @param {?} value
     * @return {?}
     */
    function (name, value) {
        return deserializeData(value, dataMappers && dataMappers[name]);
    });
    /** @type {?} */
    var emptyQueryParameters = fromPairs(oFilter.names.map((/**
     * @param {?} name
     * @return {?}
     */
    function (name) { return [name, void 0]; })));
    /** @type {?} */
    var value$ = oFilter.filterChanges;
    /** @type {?} */
    var tracker = {
        filterChanges: value$,
        paramChanges: value$.pipe(
        // there is no sense to store initial filter in query parameters
        skip(1), map((/**
         * @param {?} filter
         * @return {?}
         */
        function (filter) {
            return isSimilarFilter(filter, defaultFilter)
                // send empty set of query parameters if they are equal to default filter
                ? emptyQueryParameters
                // serialize filter
                // store whole filter under corresponding query parameters
                : fromPairs(oFilter.names
                    .map((/**
                 * @param {?} name
                 * @return {?}
                 */
                function (name) { return [name, serialize(name, filter[name])]; }))
                    // send empty query parameter if it is equal to corresponding parameter from the default filter
                    .map((/**
                 * @param {?} __0
                 * @return {?}
                 */
                function (_a) {
                    var _b = __read(_a, 2), name = _b[0], value = _b[1];
                    return [
                        name,
                        defaultFilter && isSimilarFilter(value, defaultFilter[name]) ? void 0 : value,
                    ];
                })));
        })), 
        // start with empty set of query parameters
        startWith({})),
        // take parameters from corresponding query parameters
        mergeParams: (/**
         * @param {?} params
         * @return {?}
         */
        function (params) {
            var setFilter = oFilter.setFilter;
            if (setFilter) {
                /** @type {?} */
                var filter = fromPairs(oFilter.names.map((/**
                 * @param {?} name
                 * @return {?}
                 */
                function (name) { return [name, deserialize(name, params[name])]; })));
                if (!isEmpty(filter) && !every(filter, isNil)) {
                    setFilter((/** @type {?} */ (filter)));
                }
                else if (!isNil(defaultFilter)) {
                    setFilter(defaultFilter);
                }
            }
        }),
        withDefault: (/**
         * @param {?} filter
         * @return {?}
         */
        function (filter) {
            defaultFilter = filter;
            return tracker;
        }),
    };
    return tracker;
}

export { RouteDataService, RouteTracker, controlTracker, filterTracker, filterableControl, getRouteTrackerFilter, manualTracker, observableFilterTracker, observablePlainFilterTracker };
//# sourceMappingURL=softeq-angular-route-data.js.map
