import { Router } from '@angular/router';
import { DataMapper } from '@softeq/data-mappers';
import { Observable } from 'rxjs';
import { RouteParamTracker } from './route-param-tracker';
import { RouteTracker1, RouteTracker2, RouteTracker3, RouteTracker4 } from './route-tracker';
/**
 * This service allows to request and save dynamic data associated with routes.
 *
 * Service supports two kinds of data association:
 * - data can be associated with the route by its url and query parameters (current route)
 * - data can be associated with the state id attached to the current route as a parameter (savepoint)
 *
 */
export declare class RouteDataService {
    private router;
    constructor(router: Router);
    /**
     * Creates {@link RouteTracker} using provided {@link RouteParamTracker}s as filter trackers.
     */
    tracker(): RouteTracker1<void>;
    tracker<F1>(filter: RouteParamTracker<F1>): RouteTracker1<F1>;
    tracker<F1, F2>(filter1: RouteParamTracker<F1>, filter2: RouteParamTracker<F2>): RouteTracker2<F1, F2>;
    tracker<F1, F2, F3>(filter1: RouteParamTracker<F1>, filter2: RouteParamTracker<F2>, filter3: RouteParamTracker<F3>): RouteTracker3<F1, F2, F3>;
    tracker<F1, F2, F3, F4>(filter1: RouteParamTracker<F1>, filter2: RouteParamTracker<F2>, filter3: RouteParamTracker<F3>, filter4: RouteParamTracker<F4>): RouteTracker4<F1, F2, F3, F4>;
    /**
     * Merges provided <code>data</code> into the query parameters of current route
     */
    mergeQueryParams<T>(data: T, mapper?: DataMapper<T, any>): Observable<boolean>;
}
